/*
 * nlsop.c 
 *
 * this program tries helping to understand the amorphous depuration
 * and recrystallization of SiCx while ion implantation at temperatures
 * below 400 degree celsius.
 * hopefully the program will simulate the stabilization of the
 * selforganizing lamella structure in the observed behaviour.
 *
 * refs: 
 *  - J. K. N. Lindner. Habilationsschrift, Universitaet Augsburg.
 *  - Maik Haeberlen. Diplomarbeit, Universitaet Augsburg.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "nlsop.h"

#include "dfbapi.h"
#include "random.h"

#define MAKE_AMORPH(N) *(N)|=AMORPH
#define MAKE_CRYST(N) *(N)&=~AMORPH

/* test globals - get removed or included in my_info struct later */
int amorph_count;
int cryst_count;

int usage(void)
{
 puts("usage:");
 puts("-h \t\t help");
 puts("-n \t\t no user interaction");
 puts("-Z \t\t cryst -> amorph c diffusion in z direction");
 printf("-a <value> \t slope of nuclear energy loss (default %f)\n",A_EL);
 printf("-b <value> \t nuclear energy loss offset (default %f)\n",B_EL);
 printf("-x <value> \t # x cells (default %d)\n",X);
 printf("-y <value> \t # y cells (default %d)\n",Y);
 printf("-z <value> \t # z cells (default %d)\n",Z);
 printf("-s <value> \t steps (default %d)\n",STEPS);
 printf("-d <value> \t refresh display (default %d)\n",REFRESH);
 printf("-r <value> \t amorphous influence range (default %d)\n",RANGE);
 printf("-f <value> \t pressure = <value> * 1/distance^2 (default %f)\n",A_AP);
 printf("-p <value> \t pressure offset (default %f)\n",B_AP);
 printf("-F <value> \t proportionality constant between c conc and ability to get amorphous (default %f)\n",A_CP);
 printf("-A <value> \t slope of linear c distribution (default %f)\n",A_CD);
 printf("-B <value> \t linear c distribution offset (default %f)\n",B_CD);
 printf("-D <value> \t diffusion rate from cryst to amorph cells (default %f)\n",DR_AC);
 printf("-c <value> \t diffusion rate in cryst cells (default %f)\n",DR_CC);
 printf("-e <value> \t do diffusion every <value> steps (default %d)\n",DIFF_RATE);
 printf("-W <value> \t write every <value> steps to save file (default %d)\n",RESAVE);
 puts("-C <file> \t convert file to gnuplot format");
 puts("-L <file> \t load from file");
 puts("-S <file> \t save to file");
 puts("-R <file> \t read from random file");
 puts("-P <file> \t specify implantatin profile file");
 
 return 1;
}

int process_cell(d3_lattice *d3_l,u32 x,u32 y,u32 z,info *my_info)
{
 unsigned char *thiz;
 int *conc;
 int i,j;
 int off;
 double p;

 thiz=d3_l->status+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
 conc=d3_l->extra+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
 p=my_info->b_ap*URAND_MAX;
 for(i=-(my_info->range);i<=my_info->range;i++)
 {
  for(j=-(my_info->range);j<=my_info->range;j++)
  {
   if(!(i==0 && j==0))
   {
    off=((x+d3_l->max_x+i)%d3_l->max_x)+((y+d3_l->max_y+j)%d3_l->max_x)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
    if(*(d3_l->status+off)&AMORPH) p+=my_info->a_ap*(*(d3_l->extra+off))*URAND_MAX/(i*i+j*j);
   } 
  }
 }
 p+=*conc*my_info->a_cp*URAND_MAX;
 if(!(*thiz&AMORPH))
 {
  if(get_rand(URAND_MAX)<=p)
  {
   MAKE_AMORPH(thiz);
   amorph_count++;
  }
 } else
 {
  /* assume 1-p probability */
  if(get_rand(URAND_MAX)>p)
  {
   MAKE_CRYST(thiz);
   cryst_count++;
  }
 }
 
 return 1;
}

int distrib_c(d3_lattice *d3_l,info *my_info,int step,double c_ratio)
{
 u32 x,y,z;
 int i,j,k,c;
 int offset,off;
 int carry;

 /* put one c ion somewhere in the lattice */
 if(my_info->cc<c_ratio*step)
 {
  x=get_rand(d3_l->max_x);
  y=get_rand(d3_l->max_y);
  z=get_rand_lgp(d3_l->max_z,my_info->a_cd,my_info->b_cd);
  *(d3_l->extra+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)+=1;
  (my_info->cc)++;
 }

 if(step%my_info->diff_rate==0)
 {

 for(i=0;i<d3_l->max_x;i++)
 {
  for(j=0;j<d3_l->max_y;j++)
  {
   for(k=0;k<d3_l->max_z;k++)
   {
    offset=i+j*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
    /* case amorph - amorph <-> cryst diffusion */
    if(*(d3_l->status+offset)&AMORPH)
    {
     for(c=-1;c<=1;c++)
     {
      if(c!=0)
      {
       off=((i+d3_l->max_x+c)%d3_l->max_x)+j*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
       carry=0;
       if(!(*(d3_l->status+off)&AMORPH)) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));
       if(carry!=0)
       {
        *(d3_l->extra+offset)+=carry;
        *(d3_l->extra+off)-=carry;
       }
      }
     }
     for(c=-1;c<=1;c++)
     {
      if(c!=0)
      {
       off=i+((j+c+d3_l->max_y)%d3_l->max_y)*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
       carry=0;
       if(!(*(d3_l->status+off)&AMORPH)) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));                             
       if(carry!=0)
       {
        *(d3_l->extra+offset)+=carry; 
        *(d3_l->extra+off)-=carry; 
       }
      }
     }
     if(my_info->z_diff)
     {
      for(c=-1;c<=1;c++)
      {
       if(c!=0)
       {
        off=i+j*d3_l->max_x+((k+c+d3_l->max_z)%d3_l->max_z)*d3_l->max_x*d3_l->max_y;
        carry=0;
        if(!*(d3_l->status+off)&AMORPH) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));
        if(carry!=0)
        {
         *(d3_l->extra+off)-=carry;
         *(d3_l->extra+offset)+=carry;
        }
       }
      }
     }  
    } else
    /* case not amorph - cryst <-> cryst diffusion */
    {
     for(c=-1;c<=1;c++) 
     {
      if(c!=0)
      {
       off=i+((j+c+d3_l->max_y)%d3_l->max_y)*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
       carry=0;
       if(!(*(d3_l->status+off)&AMORPH))
       {
        carry=(int)(my_info->dr_cc*(*(d3_l->extra+off)-*(d3_l->extra+offset))/2);
        if(carry!=0)
        {
         *(d3_l->extra+offset)+=carry;
         *(d3_l->extra+off)-=carry;
        }
       }
      }
     }
     for(c=-1;c<=1;c++)
     {
      if(c!=0)
      {
       off=((i+c+d3_l->max_x)%d3_l->max_x)+j*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
       carry=0;
       if(!(*(d3_l->status+off)&AMORPH))
       {
	carry=(int)(my_info->dr_cc*(*(d3_l->extra+off)-*(d3_l->extra+offset))/2);
	if(carry!=0)
        {
         *(d3_l->extra+offset)+=carry;
	 *(d3_l->extra+off)-=carry;
	}
       }
      }
     }
    }
   } /* for z */
  } /* for y */
 } /* for x */

 } /* if step modulo diff_rate == 0 */

 return 1;
}

int calc_pressure(d3_lattice *d3_l,int range)
{
 int i,j,off;
 double count,max=0;
 int x,y,z;

 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    count=0;
    for(i=-range;i<=range;i++)
    {
     for(j=-range;j<=range;j++)
     {
      if(i!=0 && j!=0)
      {
       off=((x+d3_l->max_x+i)%d3_l->max_x)+((y+d3_l->max_y+j)%d3_l->max_x)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
       if(*(d3_l->status+off)&AMORPH) count+=((double)*(d3_l->extra+off))/(i*i+j*j);
      }
     }
    }
    if(count>max) max=count;
   }
  }
 }

 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    count=0;
    for(i=-range;i<=range;i++)
    {
     for(j=-range;j<=range;j++)
     {
      if(i!=0 && j!=0)
      {
       off=((x+d3_l->max_x+i)%d3_l->max_x)+((y+d3_l->max_y+j)%d3_l->max_x)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
       if(*(d3_l->status+off)&AMORPH) count+=((double)*(d3_l->extra+off))/(i*i+j*j);
      }
     }
    }
    *(unsigned char *)(d3_l->v_ptr+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)=(unsigned char)(count*255/max);
   }
  }
 }

 return 1;
}

int calc_max_extra(d3_lattice *d3_l)
{
 int x,y,z;
 int off,max=0;

 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    off=x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
    if(*(d3_l->extra+off)>max) max=*(d3_l->extra+off);
   }
  }
 }

 return max;
}

int write_bmp(d3_lattice *d3_l,int window,u32 x,u32 y,u32 z)
{
 int fd,i,j,size=0,foo=0;
 int width=0,height=0;
 char bmpfile[MAX_CHARS];
 char buf[128];

 if(window==1)
 {
  sprintf(bmpfile,"x-z_%d.bmp",y);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_z;
  width=d3_l->max_x;
  height=d3_l->max_z;
 }
 if(window==2)
 {
  sprintf(bmpfile,"y-z_%d.bmp",x);
  foo=3*d3_l->max_y;
  size=(foo+(4-foo%4))*d3_l->max_z;
  width=d3_l->max_y;
  height=d3_l->max_z;
 }
 if(window==3)
 {
  sprintf(bmpfile,"x-y_%d.bmp",z);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_y;
  width=d3_l->max_x;
  height=d3_l->max_y;
 }

 if((fd=open(bmpfile,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open bmp file");
  return -1;
 }

 /* bmpheader */
 buf[0]='B'; /* std header start */
 buf[1]='M';
 buf[2]=(size+0x36)&0xff; /* file size */
 buf[3]=(size+0x36)>>8;
 memset(buf+4,0,6);
 buf[10]=0x36; /* offset to data */
 memset(buf+11,0,3);
 buf[14]=0x28; /* length of bmp info header */
 memset(buf+15,0,3);
 buf[18]=width&0xff; /* width and height */
 buf[19]=width>>8;
 memset(buf+20,0,2);
 buf[22]=height&0xff;
 buf[23]=height>>8;
 memset(buf+24,0,2);
 buf[26]=1; /* # planes -> 1 */
 buf[27]=0;
 buf[28]=24; /* bits per pixel -> 2^24 (true color) */
 buf[29]=0;
 memset(buf+30,0,4); /* compression -> none */
 buf[34]=size&0xff; /* data size */
 buf[35]=size>>8;
 memset(buf+36,0,2);
 buf[38]=0x12; /* res: pixel/meter */
 buf[39]=0x0b;
 memset(buf+40,0,2);
 buf[42]=0x12;
 buf[43]=0x0b;
 memset(buf+44,0,2); 
 memset(buf+46,0,8); /* no colors, no important colors */

 if(write(fd,buf,54)<54)
 {
  puts("failed writing bmp header");
  return -1;
 }
 if(window==1)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    if(*(d3_l->status+i+y*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y)&RED) memset(buf,0xff,3);
    else memset(buf,0,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   } 
  }
 }
 if(window==2)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_y;i++)
   {
    if(*(d3_l->status+x+i*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y)&RED) memset(buf,0xff,3);
    else memset(buf,0,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==3)
 {
  for(j=0;j<d3_l->max_y;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    if(*(d3_l->status+i+(d3_l->max_y-j-1)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&RED) memset(buf,0xff,3);
    else memset(buf,0,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return -1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 close(fd);

 return 1;
}

int save_to_file(char *sf,d3_lattice *d3_l,info *my_inf)
{
 int sf_fd,c;

 if((sf_fd=open(sf,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open save file");
  return -1;
 }
 if(write(sf_fd,d3_l,sizeof(d3_lattice))<sizeof(d3_lattice))
 {
  puts("failed saving d3 lattice struct");
  return -1;
 }
 if(write(sf_fd,my_inf,sizeof(info))<sizeof(info))
 {
  puts("failed saving info struct");
  return-1;
 }
 c=d3_l->max_x*d3_l->max_y*d3_l->max_z;
 if(write(sf_fd,d3_l->status,c*sizeof(unsigned char))<c*sizeof(unsigned char))
 {
  puts("failed saving status of d3 lattice sites");
  return -1;
 }
 if(write(sf_fd,d3_l->extra,c*sizeof(int))<c*sizeof(int))
 {
  puts("failed saving sites concentration");
  return -1;
 }
 close(sf_fd);

 return 1;
}

int load_from_file(char *lf,d3_lattice *d3_l,info *my_inf)
{
 int lf_fd,c;

 if((lf_fd=open(lf,O_RDONLY))<0)
 {
  puts("cannot open load file");
  return -1;
 }
 if(read(lf_fd,d3_l,sizeof(d3_lattice))<sizeof(d3_lattice))
 {
  puts("failed reading d3 lattice struct");
  return -1;
 }
 if(read(lf_fd,my_inf,sizeof(info))<sizeof(info))
 {
  puts("failed reading info struct");
  return-1;
 }
 c=d3_l->max_x*d3_l->max_y*d3_l->max_z;
 if((d3_l->status=(unsigned char*)malloc(c*sizeof(unsigned char)))==NULL)
 {
  puts("cannot allocate status buffer");
  return -1;
 }
 if((d3_l->extra=(int *)malloc(c*sizeof(int)))==NULL)
 {
  puts("cannot allocate concentration buffer");
  return -1;
 }
 if(read(lf_fd,d3_l->status,c*sizeof(unsigned char))<c*sizeof(unsigned char))
 {
  puts("failed reading status of d3 lattice sites");
  return -1;
 }
 if(read(lf_fd,d3_l->extra,c*sizeof(int))<c*sizeof(int))
 {
  puts("failed reading sites concentration");
  return -1;
 }
 close(lf_fd);

 return 1;
}

int convert_file(char *cf,d3_lattice *d3_l)
{
 int x,y,z;
 int c_fd;

 if((c_fd=open(cf,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open convert file");
  return -1;
 }
 dprintf(c_fd,"# created by nlsop (gnuplot format)\n");
 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    if(*(d3_l->status+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&AMORPH) dprintf(c_fd,"%d %d %d\n",x,y,z);
   }
  }
 }
 close(c_fd);

 return 1;
}

int get_c_ratio(double *c_ratio,char *pfile,info *my_info,d3_lattice *d3_l)
{
 double all,a,b,d;
 int i,k;
 int p_fd;
 unsigned char buf[32];
 char *p;
 unsigned char c;

 if((p_fd=open(pfile,O_RDONLY))<0)
 {
  puts("cannot open profile file");
  return -1;
 }
 k=1;
 d=0;
 all=0;
 while(k)
 {
  for(i=0;i<32;i++)
  {
   k=read(p_fd,&c,1);
   buf[i]=c;
   if(c=='\n') break;
  }
  if(k)
  {
   p=strtok(buf," ");
   a=atof(p)/10; /* nm */
   p=strtok(NULL," ");
   b=atof(p);
   if(a>my_info->b_cd*CELL_LENGTH && a<(my_info->b_cd+d3_l->max_z)*CELL_LENGTH) d+=b;
   all+=b;
  }
 }
 *c_ratio=d/all;
 close(p_fd);

 return 1;
}

int main(int argc,char **argv)
{
 u32 x,y,z,x_c,y_c,z_c;
 int i,quit,escape,switchmode,nowait,bmp;
 int refresh,resave;
 char s_file[MAX_CHARS];
 char s_file_tmp[MAX_CHARS];
 char l_file[MAX_CHARS];
 char c_file[MAX_CHARS];
 char p_file[MAX_CHARS];
 char convert;
 char r_file[MAX_CHARS];
#ifdef USE_DFB_API
 char xyz_txt[MAX_TXT];
 char status_txt[MAX_TXT];
 char conc_txt[MAX_TXT];
 char steps_txt[MAX_TXT];
 char cc_txt[MAX_TXT];
 char a_txt[MAX_TXT];
 char s_txt[MAX_TXT];
 char ap_txt[MAX_TXT];
 char el_txt[MAX_TXT];
 char cd_txt[MAX_TXT];
 char r_txt[MAX_TXT];
 char cp_txt[MAX_TXT];
 char zdiff_txt[MAX_TXT];
 char diff_txt[MAX_TXT];
 char dr_ac_txt[MAX_TXT];
 char dr_cc_txt[MAX_TXT];
 char dose_txt[MAX_TXT];
 char mode_txt[MAX_TXT];
 char *arg_v[MAX_ARGV];
#endif
 d3_lattice d3_l;
 info my_info;
 unsigned char mode;
 double c_ratio;
#ifdef USE_DFB_API
 int max_extra;
#endif

 d3_l.max_x=X;
 d3_l.max_y=Y;
 d3_l.max_z=Z;
 my_info.steps=STEPS;
 my_info.range=RANGE;
 refresh=REFRESH;
 resave=RESAVE;
 my_info.z_diff=0;
 my_info.a_el=A_EL;
 my_info.b_el=B_EL;
 my_info.a_cd=A_CD;
 my_info.b_cd=B_CD;
 my_info.a_ap=A_AP;
 my_info.b_ap=B_AP;
 my_info.a_cp=A_CP;
 my_info.cc=CC;
 my_info.dr_ac=DR_AC;
 my_info.dr_cc=DR_CC;
 my_info.diff_rate=DIFF_RATE;
 nowait=0;
 quit=0;
 escape=0;
 switchmode=0;
 strcpy(s_file,"");
 strcpy(l_file,"");
 strcpy(c_file,"");
 strcpy(p_file,IMP_PROFILE);
 convert=0;
 strcpy(r_file,"");
 mode=0;

 amorph_count=0;
 cryst_count=0;

 for(i=1;i<argc;i++)
 {
  if(argv[i][0]=='-')
  {
   switch(argv[i][1])
   {
    case 'h':
     usage();
     return -1;
    case 'n':
     nowait=1;
     break;
    case 'a':
     my_info.a_el=atof(argv[++i]);
     break;
    case 'b':
     my_info.b_el=atof(argv[++i]);
     break;
    case 'x':
     d3_l.max_x=atoi(argv[++i]);
     break;
    case 'y':
     d3_l.max_y=atoi(argv[++i]);
     break;
    case 'z':
     d3_l.max_z=atoi(argv[++i]);
     break;
    case 'Z':
     my_info.z_diff=1;
     break;
    case 's':
     my_info.steps=atoi(argv[++i]);
     break;
    case 'd':
     refresh=atoi(argv[++i]);
     break;
    case 'r':
     my_info.range=atoi(argv[++i]);
     break;
    case 'f':
     my_info.a_ap=atof(argv[++i]);
     break;
    case 'p':
     my_info.b_ap=atof(argv[++i]);
     break;
    case 'F':
     my_info.a_cp=atof(argv[++i]);
     break;
    case 'A':
     my_info.a_cd=atof(argv[++i]);
     break;
    case 'B':
     my_info.b_cd=atof(argv[++i]);
     break;
    case 'W':
     resave=atoi(argv[++i]);
     break;
    case 'C':
     strcpy(l_file,argv[++i]);
     if(i<argc-1) if(argv[i+1][0]!='-') strcpy(c_file,argv[++i]);
     convert=1;
     break;
    case 'D':
     my_info.dr_ac=atof(argv[++i]);
     break;
    case 'c':
     my_info.dr_cc=atof(argv[++i]);
     break;
    case 'e':
     my_info.diff_rate=atoi(argv[++i]);
     break;
    case 'L':
     strcpy(l_file,argv[++i]);
     break;
    case 'S':
     strcpy(s_file,argv[++i]);
     break;
    case 'R':
     strcpy(r_file,argv[++i]);
     break;
    case 'P':
     strcpy(p_file,argv[++i]);
     break;
    default:
     usage();
     return -1;
   }
  } else usage();
 }

 x=d3_l.max_x/2-1;
 y=d3_l.max_y/2-1;
 z=d3_l.max_z/2-1;

#ifdef NODFB
 if(!strcmp(s_file,""))
 {
  puts("NODFB defined, run with -S option");
  return -1;
 }
#endif

 if(!strcmp(r_file,"")) rand_init(NULL);
 else rand_init(r_file);

 if(!strcmp(l_file,""))
 {
  i=d3_l.max_x*d3_l.max_y*d3_l.max_z;
#ifdef USE_DFB_API
  d3_lattice_init(&argc,argv,&d3_l);
#endif
  if((d3_l.status=(unsigned char *)malloc(i*sizeof(unsigned char)))==NULL)
  {
   puts("failed allocating status buffer");
   return -1;
  }
  memset(d3_l.status,0,i*sizeof(unsigned char));
  if((d3_l.extra=(int *)malloc(i*sizeof(int)))==NULL)
  {
   puts("failed allocating concentration buffer");
   return -1;
  }
  memset(d3_l.extra,0,i*sizeof(int));
 } else
 {
  load_from_file(l_file,&d3_l,&my_info);
  if(convert) 
  {   
   if(!strcmp(c_file,"")) sprintf(c_file,"%s_gnuplot",l_file);
   printf("converting file %s to %s\n",l_file,c_file);
   convert_file(c_file,&d3_l);
   puts("done");
   return 1;
  } 
#ifdef USE_DFB_API
    else d3_lattice_init(&argc,argv,&d3_l);
#endif
 }

#ifdef USE_DFB_API
 d3_event_init(&d3_l);
#endif

#ifdef USE_DFB_API
 strcpy(a_txt,"args:");
 sprintf(s_txt,"steps: %d",my_info.steps);
 sprintf(dose_txt,"dose: %.2fe+17 C/cm",my_info.steps*1.0/(d3_l.max_x*d3_l.max_y*CELL_LENGTH*CELL_LENGTH*1000));
 sprintf(r_txt,"pressure range: %d",my_info.range);
 sprintf(ap_txt,"a_ap: %.3f  b_ap: %.3f",my_info.a_ap,my_info.b_ap);
 sprintf(el_txt,"a_el: %.3f  b_el: %.3f",my_info.a_el,my_info.b_el);
 sprintf(cd_txt,"a_cd: %.3f  b_cd: %.3f",my_info.a_cd,my_info.b_cd);
 sprintf(cp_txt,"a_cp: %.4f",my_info.a_cp);
 sprintf(dr_ac_txt,"a/c diffusion rate: %.4f",my_info.dr_ac);
 sprintf(dr_cc_txt,"c/c diffusion rate: %.4f",my_info.dr_cc);
 sprintf(zdiff_txt,"diffusion in z direction: %c",my_info.z_diff?'y':'n');
 sprintf(diff_txt,"diffusion every %d steps",my_info.diff_rate);
 strcpy(mode_txt,"view: a/c mode");
 arg_v[1]=xyz_txt;
 arg_v[2]=NULL;
 arg_v[3]=status_txt;
 arg_v[4]=conc_txt;
 arg_v[5]=NULL;
 arg_v[6]=mode_txt;
 arg_v[7]=NULL;
 arg_v[8]=steps_txt;
 arg_v[9]=cc_txt;
 arg_v[10]=NULL;
 arg_v[11]=diff_txt;
 arg_v[12]=zdiff_txt;
 arg_v[13]=NULL;
 arg_v[14]=a_txt;
 arg_v[15]=s_txt;
 arg_v[16]=dose_txt;
 arg_v[17]=NULL;
 arg_v[18]=r_txt;
 arg_v[19]=ap_txt;
 arg_v[20]=el_txt;
 arg_v[21]=cd_txt;
 arg_v[22]=cp_txt;
 arg_v[23]=NULL;
 arg_v[24]=dr_ac_txt;
 arg_v[25]=dr_cc_txt;
#endif

 if(!strcmp(l_file,""))
 {
  if(get_c_ratio(&c_ratio,p_file,&my_info,&d3_l)!=1)
  {
   puts("failed calculating ratio");
   return -1;
  }
  i=0;
  while((i<my_info.steps) && (quit==0) && (escape==0))
  {
   x_c=get_rand(d3_l.max_x);
   y_c=get_rand(d3_l.max_y);
   z_c=get_rand_lgp(d3_l.max_z,my_info.a_el,my_info.b_el);
   distrib_c(&d3_l,&my_info,i,c_ratio);
   process_cell(&d3_l,x_c,y_c,z_c,&my_info);
#ifdef USE_DFB_API
   if(i%refresh==0)
   {
    sprintf(xyz_txt,"x: %d  y: %d  z: %d",x+1,y+1,z+1);
    sprintf(status_txt,"status: %c",(*(d3_l.status+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y)&AMORPH)?'a':'c');
    sprintf(conc_txt,"conc: %d",*(d3_l.extra+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y));
    sprintf(steps_txt,"step: %d  a_count: %d",i,amorph_count);
    sprintf(cc_txt,"total c: %d  c_count: %d",my_info.cc,cryst_count);
    d3_lattice_draw(&d3_l,x,y,z,25,arg_v,mode,0);
   }
#endif
   if(i%resave==0 && strcmp(s_file,"") && resave!=0 && i!=0)
   {
    sprintf(s_file_tmp,"%s_%d_of_%d",s_file,i,my_info.steps);
    save_to_file(s_file_tmp,&d3_l,&my_info);
#ifdef NODFB
    printf("saved %s\n",s_file_tmp);
#endif
   }
   i++;
  }
 }

 if(strcmp(s_file,""))
 {
   printf("saved %s\n",s_file);
   save_to_file(s_file,&d3_l,&my_info);
 }

#ifdef USE_DFB_API
 /* allocating buffer for pressure values */
 if((d3_l.v_ptr=malloc(d3_l.max_x*d3_l.max_y*d3_l.max_z*sizeof(unsigned char)))==NULL)
 {
  puts("cannot allocate buffer for pressure values");
  return -1;
 }
 /* calc values */
 calc_pressure(&d3_l,my_info.range);
 max_extra=calc_max_extra(&d3_l);

 while((quit==0) && (escape==0) && (nowait==0))
 {
  /* bahh! */
  if(switchmode==0) mode=0;
  if(switchmode==1) mode=1;
  if(switchmode==2) mode=2;
  /* end of bahh! */
  sprintf(xyz_txt,"x: %d  y: %d  z: %d",x+1,y+1,z+1);
  sprintf(status_txt,"status: %c",(*(d3_l.status+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y)&AMORPH)?'a':'c');
  sprintf(conc_txt,"conc: %d",*(d3_l.extra+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y));
  sprintf(steps_txt,"step: end");
  sprintf(cc_txt,"total c: %d",my_info.cc);
  if(switchmode==0) strcpy(mode_txt,"view: a/c mode");
  if(switchmode==1) strcpy(mode_txt,"view: c conc mode");
  if(switchmode==2) strcpy(mode_txt,"view: a pressure mode");
  d3_lattice_draw(&d3_l,x,y,z,25,arg_v,mode,max_extra);
  bmp=0;
  scan_event(&d3_l,&x,&y,&z,&quit,&escape,&switchmode,&bmp);
  if(bmp) write_bmp(&d3_l,bmp,x,y,z);

 }

 d3_lattice_release(&d3_l);
#endif

 return 1;
}
